// Beispiel für Default-Datentyp in Templates
// Realisiert ein dynamisches Feld
// ACHTUNG!! Werden in dem Feld Objektzeiger abgelegt
// so müssen die dazugehörigen Objekte explizit wieder
// zerstört werden.

#include <iostream>
#include <type_traits>

// Demo-Klasse einbinden
#include "demo.h"

using std::cout;
using std::endl;

// Klasse implementiert einen gesicherten Zugriff auf
// ein Feld. Wird beim Schreiben über das Ende des Feldes
// hinausgegriffen, so wird das Feld entsprechend erweitert
// Standardmässig wird das Feld für die Ablage von int-
// Daten angelegt
template <typename T = int>
class DynArray
{
    T* data; // Zeiger auf Datenfeld
    size_t size; // Grösse des Datenfelds
  public:
    // ctor, reserviert optional ein Feld mit einer
    // bestimmten Anzahl von Elementen
    DynArray(decltype(size) reserveData = 0);
    // dtor
    ~DynArray();
    // Indexoperator für Array-Zugriff
    T& operator[] (decltype(size) index);
    // Groesse des Feldes zurueckgeben
    size_t GetSize();
    // Ausgabe des DynArrays
    void PrintArray();
};
// ctor
// Standardmässig wird kein Feld allokiert und der Datenzeiger
// mit nullptr initialisiert
template <typename T>
DynArray<T>::DynArray(decltype(size) reserveData): data(nullptr), size(0)
{
    // Falls Datenfeld angelegt werden soll
    if (reserveData)
    {
        // Datenfeld reservieren
        data = new T[reserveData];
        size = reserveData;
        if (std::is_fundamental<T>::value)
        {
            for (size_t toInit=0; toInit<size; toInit++)
                data[toInit] = 0;
        }
    }
}
// dtor, gibt reserviertes Feld frei
template <typename T>
DynArray<T>::~DynArray()
{
    delete [] data;
}
// Überladener Indexoperator für den Zugriff
// auf die Elemente im DynArray
template <typename T>
T& DynArray<T>::operator[] (decltype(size) index)
{
    // Falls Index grösser als die Anzahl der Feldelement
    if (index>=size)
    {
        // Neues Feld mit der Grösse 'Index+1' reservieren
        // +1 hier unbedingt notwendig, da z.B. der
        // Index 10 das 11. Element adressiert!
        auto newData = new T[index+1];
        // Bisherige Daten in neues Feld übernehmen
        // Move-Operator um unnötige Kopierarbeite zu sparen!
        for (size_t copyOld=0; copyOld<size; copyOld++)
            newData[copyOld] = std::move(data[copyOld]);
        if (std::is_fundamental<T>::value)
        {
            for (auto toInit=size; toInit<index; toInit++)
                newData[toInit] = 0;
        }
        // Altes Feld nun entfernen
        delete [] data;
        // Und Zeige auf neues Feld abspeichern
        data = newData;
        // Nicht vergessen: die Grösse korrigieren
        size = index+1;
    }
    // Gewünschtes Element zurückliefern
    return data[index];
}
template <typename T>
inline size_t DynArray<T>::GetSize()
{
    return size;
}
// Komplettes Feld ausgeben
template <typename T>
void DynArray<T>::PrintArray()
{
    for (decltype(size) index=0; index<size; index++)
        cout << data[index] << ", ";
    cout << endl;
}

// main() Funktion
int main()
{
    // DynArray mit Default-Datentyp 'int'
    DynArray<> intArray;
    // 5. Element beschreiben
    intArray[4] = 44;
    cout << "Initiale int-Feldgroesse: " << intArray.GetSize() << endl;
    intArray.PrintArray();
    // 7. Element beschreiben -> führt zur Erweiterung des DynArrays
    intArray[6] = 66;
    // Komplettes Feld ausgeben (7 Elemente)
    cout << "Aktuelle int-Feldgroesse: " << intArray.GetSize() << endl;
    cout << "Und das komplette int-Feld:\n";
    intArray.PrintArray();

    // DynArray mit 3 Demo-Objekten anlegen
    DynArray<Demo> objArray(3);
    // 2. Element beschreiben
    objArray[1] = Demo("Objekt 1");
    // Komplettes Feld ausgeben (3 Elemente)
    cout << "Initiale Demo-Feldgroesse: " << objArray.GetSize() << endl;
    objArray.PrintArray();
    // 6. Element beschreiben -> führt zur Erweiterung des DynArrays
    objArray[5] = Demo("Objekt 5");
    // Komplettes Feld ausgeben (6 Elemente)
    cout << "Aktuelle Demo-Feldgroesse: " << objArray.GetSize() << endl;
    cout << "Und das komplette Demo-Feld:\n";
    objArray.PrintArray();
}
