#include <iostream>
#include <string>
#include <cstdlib>
#include "demo.h"

using std::cout;
using std::endl;
using std::string;

// Templates zur Reservierung von Speicher.
// Templates werden nachher als Parameter eines weiteren
// Templates verwendet um die Art der Speicherreservierung
// zu bestimmen
// Hinweis: Die Templates sind jetzt vom Typ struct da
// alle Memberfunktionen public sind
// 1. Speicherreservierung mittels new
template <class T>
struct NewAlloc
{
    static T* Create()
    {
        cout << "NewAlloc::Create()" << endl;
        return new T;
    }
    static void Release(T *ptr)
    {
        cout << "NewAlloc::Release()" << endl;
        delete ptr;
    }
};
// 2. Speicherreservierung mittels malloc
// In der Praxis könnte z.B anstelle von malloc() die
// Reservierung von Speicher in einem bestimmten Bereich
// (z.B. shared-Memory) erfolgen
template <class T>
struct MallocAlloc
{
    static T* Create()
    {
        cout << "MallocAlloc::Create()" << endl;
        void *pMem = std::malloc(sizeof(T));
        // Wenn mit malloc Speicher fuer Objekte reserviert
        // wird, so muss der ctor des Objekts explizit mit
        // dem placement-new Operator aufgerufen werden
        if (pMem != NULL)
           return new(pMem) T;
        return NULL;
    }
    static void Release(T* ptr)
    {
        cout << "MallocAlloc::Release()" << endl;
        // Explizter Aufruf des dtors erforderlich!
        ptr->~T();
        free(ptr);
    }
};

// Template zum Abspeichern von Daten
// Template erhaelt als ersten Parameter den Datentyp des Datums
// und als zweiten Parameter das Template zur Reservierung des
// Speichers. Das Template zur Speicherreservierung erhaelt seinen
// Template-Parameter automatisch aus dem Datentyp-Parameter
template <typename T, template <typename> class TAlloc=NewAlloc >
class StoreData
{
    T* pData; // Zeiger auf das abzulegende Datum
  public:
    StoreData()
    {
        // Aufruf der entsprechenden Create()-Memberfunktion
        // TAlloc ist Klassentemplate!
        pData = TAlloc<T>::Create();
    }
    ~StoreData()
    {
        // Aufruf der entsprechenden Destroy() Memberfunktion
        TAlloc<T>::Release(pData);
    }
    // Datum setzen
    void SetData(const T& data)
    {
        *pData = data;
    }
    // Datum ausgeben
    void PrintData() const
    {
        cout << "Datum: " << *pData << endl;
    }
};

// main() Funktion
int main()
{
    // Platz fuer int-Wert mit malloc anlegen
    cout << "int mit MallocAlloc:\n";
    StoreData<int, MallocAlloc> intParam;
    intParam.SetData(10);
    intParam.PrintData();
    // Platz fuer string-Wert mit new anlegen
    cout << "\nstring mit NewAlloc:\n";
    StoreData<string, NewAlloc> stringParam;
    stringParam.SetData("Another string");
    stringParam.PrintData();
    // Platz fuer float-Wert mit Default-Allokierung anlegen
    cout << "\nfloat mit Default-Template:\n";
    StoreData<float> floatParam;
    floatParam.SetData(3.1416f);
    floatParam.PrintData();
    // Platz fuer Demo-Objekt mit malloc anlegen
    cout << "\nDemo mit MallocAlloc:\n";
    StoreData<Demo, MallocAlloc> classParam;
    classParam.PrintData();
    classParam.SetData(Demo("A new string"));
    classParam.PrintData();
    // Ab hier wird wieder aufgeräumt
    cout << "\nUnd alles wieder entfernen:\n";
}
